#ifndef LEVEL_INTERFACE_NETWORK_WRITER
#define LEVEL_INTERFACE_NETWORK_WRITER

#include <string>
#include <dune/common/fvector.hh>
#include <dune/fufem/boundaryiterator.hh>
//#include <dune/faultnetworks/levelinterfacenetwork.hh>

class LevelInterfaceNetworkWriter {
    private:
        std::string filePath_;

        template <typename Intersection>
        void writeIntersection(const Intersection& intersection, std::ofstream& file, const int colorID) {
            typedef typename Dune::FieldVector<typename Intersection::ctype, Intersection::dimensionworld > GlobalCoordinates;

            const auto& geometry = intersection.geometry();

            std::vector<GlobalCoordinates> vertices;
            vertices.resize(geometry.corners());

            for (int i=0; i<geometry.corners(); i++) {
                const auto& vertex = geometry.corner(i);
                vertices[i] = vertex;
            }

            for (size_t i=0; i<vertices.size(); i++) {
                for (size_t j=i+1; j<vertices.size(); j++) {

                    switch (colorID) {
                        case 0:
                            file << "\\draw[black] ";
                            break;
                        case 1:
                            file << "\\draw[MATLABblue] ";
                            break;
                        case 2:
                            file << "\\draw[MATLABgreen] ";
                            break;
                        case 3:
                            file << "\\draw[MATLABred] ";
                            break;
                        case 4:
                            file << "\\draw[MATLAByellow] ";
                            break;
                        default:
                            file << "\\draw[black] ";
                    }

                    writeVertex(vertices[i], file);
                    file << " -- ";
                    writeVertex(vertices[j], file);
                    file << ";\n";
                }
            }
        }

        template <typename GlobalCoordinates>
        void writeVertex(const GlobalCoordinates& vertex, std::ofstream& file) {
            size_t lastIndex = vertex.size()-1;

            file << "(";
            for (size_t i=0; i<lastIndex; i++) {
                file << vertex[i] << ", ";
            }
            file << vertex[lastIndex] << ")";
        }

    public:
        LevelInterfaceNetworkWriter(const std::string filePath) :
            filePath_(filePath)
        {}

        void setTarget(const std::string filePath) {
            filePath_ = filePath;
        }

        // works only in 1D and 2D
        template <typename LevelInterfaceNetwork>
        void write(const LevelInterfaceNetwork& levelInterfaceNetwork, bool highlightNewLevelInterfaces = false) {
			typedef typename LevelInterfaceNetwork::GridType::LevelGridView GridView;
            typedef typename LevelInterfaceNetwork::Intersection Intersection;

            std::ofstream file (filePath_);

            if (file.is_open()) {
                // define custom MATLAB colors
                file << "\\definecolor{MATLABred}{RGB}{162,20,47}\n";
                file << "\\definecolor{MATLABblue}{RGB}{0,114,189}\n";
                file << "\\definecolor{MATLABgreen}{RGB}{60,140,40}\n";
                file << "\\definecolor{MATLABorange}{RGB}{212,83,25}\n";
                file << "\\definecolor{MATLAByellow}{RGB}{237,177,32}\n";

                file << "\\begin{tikzpicture}[scale=\\scale]\n";

                const GridView& gridView = levelInterfaceNetwork.levelGridView();

                // domain boundary
                BoundaryIterator<GridView> bIt(gridView, BoundaryIterator<GridView>::begin);
                BoundaryIterator<GridView> bEnd(gridView, BoundaryIterator<GridView>::end);
				for(; bIt!=bEnd; ++bIt) {
                    writeIntersection(*bIt, file, 0);
				}

				// fault intersections
                const std::vector<Intersection>& intersections = levelInterfaceNetwork.getIntersections();
                const std::vector<int>& intersectionsLevels = levelInterfaceNetwork.getIntersectionsLevels();
                for (size_t i=0; i<intersections.size(); i++) {
                    if (highlightNewLevelInterfaces) {
                        writeIntersection(intersections[i], file, 3*(intersectionsLevels[i]==levelInterfaceNetwork.level()));
                    } else {
                        writeIntersection(intersections[i], file, intersectionsLevels[i]);
                    }
                }

                file << "\\end{tikzpicture}\n";
                file.close();
            } else {
                DUNE_THROW(Exception, "Unable to open " << filePath_ << " for writing!");
            }
        }
};
#endif
